#*******************************************************************************
#**  This file is part of Dux Rei Nummariae.                                 ***
#**                                                                          ***
#**  Copyright (C) 2017, 2018, 2019, 2020, 2021                              ***
#**  John Schneiderman <Licensing _AT_ Schneiderman _DOT_ me>                ***
#**                                                                          ***
#**  This program is free software: you can redistribute it and/or modify it ***
#**  under the terms of the GNU Affero General Public License as published   ***
#**  by the Free Software Foundation, either version 3 of the License, or    ***
#**  (at your option) any later version.                                     ***
#**                                                                          ***
#**  This program is distributed in the hope that it will be useful, but     ***
#**  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
#**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
#**  See the GNU Affero General Public License for more details.             ***
#**                                                                          ***
#**  You should have received a copy of the GNU Affero General Public License***
#**  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
#*******************************************************************************
macro (installAnyQtDependencies target_file_name)
	# TODO: Add support for GNU/Linux and macOS.
	# TODO: Would be nice to have it run within the Visual Studio Command Prompt.
	if (WIN32)
	install(
		CODE
		"
		execute_process(
			COMMAND \"${CMAKE_COMMAND}\" -E
				env PATH=\"${QT_BIN_DIRECTORY}\" \"${DEPLOYQT_EXECUTABLE}\"
					--no-compiler-runtime
					\"${CMAKE_INSTALL_PREFIX}/bin/${target_file_name}\"
			COMMAND_ERROR_IS_FATAL ANY
		)
		"
		COMPONENT app
	)
	endif()
endmacro()

macro (setUpProject projectType)
	if (NOT DEFINED ${PROJECT_NAME}_PUBLIC_DECLARATIONS)
		message(
			FATAL_ERROR
			"The project '${PROJECT_NAME}' is missing the public declaration variable."
		)
	endif (NOT DEFINED ${PROJECT_NAME}_PUBLIC_DECLARATIONS)

	if (NOT DEFINED ${PROJECT_NAME}_PUBLIC_DEFINITIONS)
		message(
			FATAL_ERROR
			"The project '${PROJECT_NAME}' is missing the public definition variable."
		)
	endif (NOT DEFINED ${PROJECT_NAME}_PUBLIC_DEFINITIONS)

	if (Qt5_FOUND)
		get_target_property(QMAKE_EXECUTABLE Qt5::qmake IMPORTED_LOCATION)
		get_filename_component(QT_BIN_DIRECTORY "${QMAKE_EXECUTABLE}" DIRECTORY)
		find_program(
			DEPLOYQT_EXECUTABLE
			NAMES windeployqt macdeployqt
			HINTS "${QT_BIN_DIRECTORY}"
			DOC "The Qt deployment binary path and file-name"
		)
	endif ()

	source_group("External Files"
		FILES ${${PROJECT_NAME}_PUBLIC_DECLARATIONS} ${${PROJECT_NAME}_PUBLIC_DEFINITIONS}
	)
	source_group("Internal Files"
		FILES ${${PROJECT_NAME}_PRIVATE_DECLARATIONS} ${${PROJECT_NAME}_PRIVATE_DEFINITIONS}
	)

	add_definitions(-DQT_NO_NARROWING_CONVERSIONS_IN_CONNECT)

	if ("${projectType}" STREQUAL "DynamicLibrary")
		if (NOT DEFINED ${PROJECT_NAME}_PRIVATE_DECLARATIONS)
			message(
				FATAL_ERROR
				"The project '${PROJECT_NAME}' is missing the private declaration variable."
			)
		endif (NOT DEFINED ${PROJECT_NAME}_PRIVATE_DECLARATIONS)

		if (NOT DEFINED ${PROJECT_NAME}_PRIVATE_DEFINITIONS)
			message(
				FATAL_ERROR
				"The project '${PROJECT_NAME}' is missing the private definition variable."
			)
		endif (NOT DEFINED ${PROJECT_NAME}_PRIVATE_DEFINITIONS)
		add_library(${PROJECT_NAME} SHARED
			"${PROJECT_BINARY_DIR}/${PROJECT_NAME}_export.h"
			${${PROJECT_NAME}_PUBLIC_DEFINITIONS}
			${${PROJECT_NAME}_PRIVATE_DEFINITIONS}
			${${PROJECT_NAME}_PUBLIC_DECLARATIONS}
			${${PROJECT_NAME}_PRIVATE_DECLARATIONS}
		)
		target_include_directories(${PROJECT_NAME}
			PUBLIC
				# Needed for internal projects of the library to be able to access the project code.
				$<BUILD_INTERFACE:${CMAKE_CURRENT_SOURCE_DIR}/external>
				# Needed for the internal projects be able to access the auto-generated code, e.g.
				# the export file.
				$<BUILD_INTERFACE:${PROJECT_BINARY_DIR}>
			PRIVATE
				# Needed for the library to be able to access the internal code.
				$<BUILD_INTERFACE:${CMAKE_CURRENT_SOURCE_DIR}>
				# Needed for the library be able to access the auto-generated code, e.g.
				# the export file.
				$<BUILD_INTERFACE:${PROJECT_BINARY_DIR}>
				# Needed for the auto-generated code at the top level, e.g. config.h.
				$<BUILD_INTERFACE:${CMAKE_BINARY_DIR}>
		)
		set_target_properties(${PROJECT_NAME} PROPERTIES
			VERSION ${${CMAKE_PROJECT_NAME}_VERSION_MAJOR}.${${CMAKE_PROJECT_NAME}_VERSION_MINOR}.${${CMAKE_PROJECT_NAME}_VERSION_PATCH}
			SOVERSION ${${CMAKE_PROJECT_NAME}_VERSION_MAJOR}
			LINKER_LANGUAGE CXX
			COMPATIBLE_INTERFACE_STRING ${PROJECT_VERSION_MAJOR}
		)
		set_property(TARGET ${PROJECT_NAME} PROPERTY INTERFACE_${PROJECT_NAME}_MAJOR_VERSION ${PROJECT_VERSION_MAJOR})
		include(GenerateExportHeader)
		generate_export_header(${PROJECT_NAME}
			PREFIX_NAME DRN_
		)

		if (${CMAKE_PROJECT_NAME}_ENABLE_TESTS)
			message("Adding unit-tests for project ${PROJECT_NAME}.")
			add_subdirectory(unit-tests)
		endif (${CMAKE_PROJECT_NAME}_ENABLE_TESTS)
		install(
			TARGETS ${PROJECT_NAME}
			RUNTIME DESTINATION bin
			COMPONENT app
			LIBRARY DESTINATION ${DuxReiNummariae_LIBRARY_INSTALL_DIR}
			COMPONENT dev
			ARCHIVE DESTINATION ${DuxReiNummariae_LIBRARY_INSTALL_DIR}
			COMPONENT dev
			INCLUDES DESTINATION include
			COMPONENT dev
		)
		install(
			FILES ${${PROJECT_NAME}_PUBLIC_DECLARATIONS} "${PROJECT_BINARY_DIR}/${PROJECT_NAME}_export.h"
			DESTINATION include/drn/${PROJECT_NAME}
			COMPONENT dev
		)

		if (WIN32)
			installAnyQtDependencies(${PROJECT_NAME}.dll)
			install(
				FILES $<TARGET_PDB_FILE:${PROJECT_NAME}>
				DESTINATION bin
				COMPONENT debug
			)
		endif(WIN32)
	elseif ("${projectType}" STREQUAL "UnitTest")
		foreach (project_definition IN LISTS ${PROJECT_NAME}_PUBLIC_DECLARATIONS)
			if ("${project_definition}" MATCHES ".*\\-unit\\-tests\\.h|hpp;?")
				message(
					FATAL_ERROR
					"The project '${PROJECT_NAME}' is a unit-test and cannot have test declaration file '${project_definition}'."
				)
			endif ("${project_definition}" MATCHES ".*\\-unit\\-tests\\.h|hpp;?")
		endforeach(project_definition)

		foreach (project_definition IN LISTS ${PROJECT_NAME}_PUBLIC_DEFINITIONS)
			if ("${project_definition}" MATCHES ".*\\-unit\\-tests\\.cpp;?")
				message(
					FATAL_ERROR
					"The project '${PROJECT_NAME}' is a unit-test and cannot have test definitions file '${project_definition}'."
				)
			endif ("${project_definition}" MATCHES ".*\\-unit\\-tests\\.cpp;?")
		endforeach(project_definition)

		if (DEFINED ${PROJECT_NAME}_PRIVATE_DECLARATIONS)
			message(
				FATAL_ERROR
				"The project '${PROJECT_NAME}' is a unit-test and cannot have private declarations."
			)
		endif (DEFINED ${PROJECT_NAME}_PRIVATE_DECLARATIONS)

		if (DEFINED ${PROJECT_NAME}_PRIVATE_DEFINITIONS)
			message(
				FATAL_ERROR
				"The project '${PROJECT_NAME}' is a unit-test and cannot have private definitions."
			)
		endif (DEFINED ${PROJECT_NAME}_PRIVATE_DEFINITIONS)
	elseif ("${projectType}" STREQUAL "DesktopApplication")
		if (DEFINED ${PROJECT_NAME}_PRIVATE_DECLARATIONS)
			message(
				FATAL_ERROR
				"The project '${PROJECT_NAME}' is an application and cannot have private declarations."
			)
		endif (DEFINED ${PROJECT_NAME}_PRIVATE_DECLARATIONS)

		if (DEFINED ${PROJECT_NAME}_PRIVATE_DEFINITIONS)
			message(
				FATAL_ERROR
				"The project '${PROJECT_NAME}' is an application and cannot have private definitions."
			)
		endif (DEFINED ${PROJECT_NAME}_PRIVATE_DEFINITIONS)

		add_executable(${PROJECT_NAME} WIN32 MACOSX_BUNDLE ${${PROJECT_NAME}_PUBLIC_DEFINITIONS})
		set_target_properties(${PROJECT_NAME} PROPERTIES
			VERSION ${${CMAKE_PROJECT_NAME}_VERSION_MAJOR}.${${CMAKE_PROJECT_NAME}_VERSION_MINOR}.${${CMAKE_PROJECT_NAME}_VERSION_PATCH}
		)
		target_include_directories(${PROJECT_NAME}
			PRIVATE
				# Needed for the auto-generated code at the top level, e.g. config.h.
				$<BUILD_INTERFACE:${CMAKE_BINARY_DIR}>
		)
		install(
			TARGETS ${PROJECT_NAME}
			RUNTIME DESTINATION bin
			COMPONENT app
			INCLUDES DESTINATION include
			COMPONENT dev
		)
		install(
			FILES ${${PROJECT_NAME}_PUBLIC_DECLARATIONS}
			DESTINATION include/drn/${PROJECT_NAME}
			COMPONENT dev
		)

		if (WIN32)
			configure_file(
				"${CMAKE_SOURCE_DIR}/src/project.vcxproj.user.in"
				${PROJECT_BINARY_DIR}/${PROJECT_NAME}.vcxproj.user @ONLY
			)
			install(
				FILES $<TARGET_PDB_FILE:${PROJECT_NAME}>
				DESTINATION bin
				COMPONENT debug
			)

			#{ Install 3rd party libraries needed

			installAnyQtDependencies(${PROJECT_NAME}.exe)
			install(
				FILES "${pecunia_DIR}/../../../bin/pecunia.dll"
				DESTINATION bin
				COMPONENT app
			)
			install(
				FILES "${pecunia_DIR}/../../../bin/pecunia.pdb"
				DESTINATION bin
				COMPONENT debug
				OPTIONAL
			)
			install(
				FILES "${pecunia_DIR}/../../../lib/pecunia.lib"
				DESTINATION lib
				COMPONENT dev
			)

			#}
		endif(WIN32)

		if (NOT TARGET uninstall)
			configure_file(
				"${CMAKE_CURRENT_SOURCE_DIR}/cmake_uninstall.cmake.in"
				"${CMAKE_CURRENT_BINARY_DIR}/cmake_uninstall.cmake"
				IMMEDIATE @ONLY
			)

			add_custom_target(
				uninstall
				COMMAND ${CMAKE_COMMAND} -P ${CMAKE_CURRENT_BINARY_DIR}/cmake_uninstall.cmake
			)
		endif()

		set(CPACK_PACKAGE_VENDOR "John M. Schneiderman")
		set(CPACK_PACKAGE_DESCRIPTION_SUMMARY ${DuxReiNummariae_DESCRIPTION})
		set(CPACK_PACKAGE_INSTALL_DIRECTORY "Dux Rei Nummariae")
		set(CPACK_RESOURCE_FILE_LICENSE "${CMAKE_SOURCE_DIR}/COPYING")
		# TODO: The script files should be generated per executable and library project, for now it must be done by hand.
		if (WIN32)
		set(CPACK_POST_BUILD_SCRIPTS "${CMAKE_SOURCE_DIR}/src/package.cmake")
		endif ()

		include(CPack)
	else ()
		message(FATAL_ERROR "Unknown project type supplied, '${projectType}'.")
	endif ("${projectType}" STREQUAL "DynamicLibrary")
endmacro (setUpProject)

macro (addTest test_name link_library_names)
	set(test_project_name "${test_name}-tests")
	set(test_file_name "${test_name}-unit-tests")
	# This must set must be done since we're inside a macro.
	set(test_link_library_names ${link_library_names} ${ARGN})

	#{ Figure out the name of the library being tested

	add_executable(${test_project_name} ${test_file_name}.cpp ${${PROJECT_NAME}_PUBLIC_DEFINITIONS})
	get_filename_component(parent_project "${PROJECT_SOURCE_DIR}/.." DIRECTORY)
	get_filename_component(parent_project "${PROJECT_SOURCE_DIR}/.." ABSOLUTE)
	string(REPLACE "/" ";" split_parent_project "${parent_project}")
	list(GET split_parent_project -1 parent_project)

	#}

	target_include_directories(${test_project_name}
		PRIVATE
			# Gives access to test internal library code.
			$<BUILD_INTERFACE:${PROJECT_SOURCE_DIR}>/..
			# Gives access to the auto-generated internal library code.
			$<BUILD_INTERFACE:${PROJECT_BINARY_DIR}>/../${parent_project}_autogen/include
	)
	target_link_libraries(${test_project_name} PRIVATE Qt5::Core)
	target_link_libraries(${test_project_name} PRIVATE Qt5::Test)

	foreach (library IN LISTS test_link_library_names)
		target_link_libraries(${test_project_name} PRIVATE ${library})
	endforeach(library)
	add_test(
		NAME ${test_name}
		COMMAND "${test_name}-tests"
		WORKING_DIRECTORY "${CMAKE_CURRENT_BINARY_DIR}"
	)

	if (WIN32)
		configure_file(
			"${CMAKE_SOURCE_DIR}/src/project.vcxproj.user.in"
			${PROJECT_BINARY_DIR}/${test_project_name}.vcxproj.user @ONLY
		)
	endif(WIN32)
	# Clean-up macro variables.
	unset(test_project_name)
	unset(test_file_name)
	unset(test_link_library_names)
	unset(parent_project)
	unset(split_parent_project)
endmacro (addTest)
